//			dist2whole.C
//
//		complete_distribute_elements()
//
// written by Zheng gengbin

/*
This pass complete the distribute direvtives. Add default format or processor
if they missed .
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "pass.h"
#include "ext_func.h"

extern int num_procs,num_dims;

// get the processor (already defined )whose dimension is dim0 , and size is size0
SgSymbol *getProc(SgStatement *hedr_stmt , int dim0 ,int size0)
{
//printf("dim0:%d , size0:%d\n",dim0,size0);
	int i;
	for (i=0 ; i< hedr_stmt->numberOfAttributes(Proc_Def) ; i++)
	{
		SgSymbol *proc = (SgSymbol *)hedr_stmt->attributeValue(i,Proc_Def);
		int dim  = ((SgArrayType*)proc->type())->dimension();
		int size = ((SgArrayType*)proc->type())->getSize();
//printf("proc:%s,dim:%d , size:%d\n",proc->identifier(),dim,size);
		if ( dim == dim0  && size == size0 ) return proc;
	        else {
			printf("WARNING:: MORE THAN ONE PROCESSORS ARE DECLARED, Aborted.\n");
			exit(1);
		}
	}

	  // then search globally in other scope.
	SgStatement *global = hedr_stmt->controlParent();
        while (global->controlParent()) global = global->controlParent();
        ASSERT(global);
	for (i=0 ; i< global->numberOfAttributes(Proc_Def) ; i++)
	{
		SgSymbol *proc = (SgSymbol *)global->attributeValue(i,Proc_Def);
		SgType *type = proc->type();
		int dim  = ((SgArrayType*)proc->type())->dimension();
		int size = ((SgArrayType*)proc->type())->getSize();
//printf("proc:%s,dim:%d , size:%d\n",proc->identifier(),dim,size);
		if ( dim == dim0  && size == size0 ) {
			  // then try to replicate the processor's declaration.
			SgSymbol *newproc = &proc->copy();
			newproc->setScope(*hedr_stmt);
			SgArrayType *arraytype = isSgArrayType(type);
			SgExpression *ref;
			if ( arraytype) {
				ref = new SgArrayRefExp(*newproc);
				isSgArrayRefExp(ref)->setSubscripts(*arraytype->getDimList()->copyPtr());
			}
			else
				ref = new SgVarRefExp(newproc);

			SgStatement *decl=new SgVarDeclStmt(*ref, *type);
        		decl->setVariant(PROCESSORS_STMT);
        		hedr_stmt->firstStmtForDeclaration()->insertStmtAfter(*decl, *hedr_stmt);
	  		  // add attribute of Proc_Def
			hedr_stmt->addAttribute(Proc_Def , newproc , sizeof(newproc) );
	  		  // add attribute to global_bfnd also
			global->addAttribute(Proc_Def , newproc , sizeof(newproc) );
			return newproc;
		}
	        else {
			printf("WARNING:: MORE THAN ONE PROCESSORS ARE DECLARED, Aborted.\n");
			exit(1);
		}
	}
	return NULL;
}

// pow() always cause problem when link(-lm) , so write one myself.
inline int pow0(int x,int y)
{
        return y==0? 1 : x*pow0(x,y-1);
}

int devide_dim_proc(int nproc,int dim , int (&dimarray)[8])
{
	int i;

	if (dim == 0)   return 0;
	if (nproc < 2) return 0;
	if (dim == 1)  { dimarray[0] = nproc ; return 1; }

	// find a better way to divide balantly : pow(i,dim) = nproc
	for (i=1; i<=nproc/2 ; i++)
		if (pow0(i,dim) == nproc)
		{
			for (int j = 0; j<dim ; j++ ) dimarray[j] = i;
			return 1;
		}
	// so have to try one by one
	for (i = 2; i<nproc ; i++)
	{
		if (nproc % i != 0) continue;
		dimarray[dim-1]=i; 
		if (!devide_dim_proc(nproc/i,dim-1,dimarray)) continue;
		else break;
	}
	if (i==nproc) return 0;
	else return 1;
}

SgSymbol  *get_dim_processor(SgStatement *hedr_stmt,int dim)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR  || variant == MODULE_STMT);

	// find an already defined one
	SgSymbol *proc = getProc(hedr_stmt, dim , num_procs);
	if (proc) return proc;

	proc = get_temp_variable(hedr_stmt,"PROC");
	SgType   *type = SgTypeProcessors();
	int       dimarray[8];

  
	if (num_procs != -1 && !devide_dim_proc(num_procs,dim,dimarray))
		return NULL;

	SgExpression *expr_list=NULL;
	  // call number_of_processors to decide processor number in runtime
	if (num_procs == -1) {
	    //  processor is decalred as P(number_of_processors(), 1, 1, ...)
	  SgSymbol *symb = hedr_stmt->getSymbolInCurrentScope("number_of_processors");
	  SgFunctionCallExp *fun = new SgIntrinsicFunctionCallExp(*symb);
 	  for (int i=0; i<dim; i++) {
            SgExpression *ddot;	
	    if (i != dim-1)
	  	ddot=new SgSubscriptExp(*new SgValueExp(1),*new SgValueExp(1));
	    else
	  	ddot=new SgSubscriptExp(*new SgValueExp(1),*fun);
	    expr_list = new SgExpression(EXPR_LIST,ddot,expr_list,NULL);
	  }
	}
	else
	  for (int i =0; i< dim ; i++)
	  {
		SgExpression *ddot=new SgSubscriptExp(*new SgValueExp(1),*new SgValueExp(dimarray[i]));
		expr_list = new SgExpression(EXPR_LIST,ddot,expr_list,NULL);
	  }
	SgArrayType *arraytype=new SgArrayType(*type,dim,*expr_list);
	proc->setType(arraytype);

	SgExpression *arrayref=new SgArrayRefExp(*proc,*expr_list);
	SgExpression *ref = arrayref;

	SgStatement *decl=new SgVarDeclStmt(* ref,* type);
        decl->setVariant(PROCESSORS_STMT);
        hedr_stmt->firstStmtForDeclaration()->insertStmtAfter(*decl);
	  // add attribute of Proc_Def
	hedr_stmt->addAttribute(Proc_Def , proc , sizeof(proc) );
	  // add attribute to global_bfnd also
        SgStatement *global = hedr_stmt->controlParent();
	while (global->controlParent()) global = global->controlParent();
	ASSERT(global);
	global->addAttribute(Proc_Def , proc , sizeof(proc) );
	
//	defaultProcs[dim]=proc;
	return proc;
}

void complete_distribute_elements(SgStatement *s)
{
	SgExpression *dist_op , *dist_format , *dist_target;
	SgSymbol     *distributee;

	dist_op = s->expr(2)->lhs();
	dist_format = dist_op->lhs();
	dist_target = dist_op->rhs();
	distributee = s->expr(0)->lhs()->symbol();
//printf("distributee:%s\n",distributee->identifier());

	// if the distribute format is missed
	if (dist_format == NULL || dist_format->rhs() == NULL)
	{
	    if (distributee->type()->variant() == T_ARRAY) 
	    {
		SgArrayType  *atype = (SgArrayType *)(distributee->type());
		int d;	// dim of processors
		if ( num_dims == -1)
			d = atype->dimension();
		else
			d = num_dims;		// if defined in argv
		SgExpression *format_list = NULL;
		for (int i=0; i<atype->dimension() ; i++ )
		{
			SgExpression *node;
#ifdef DEFAULT_COLLPASE
			node= new SgExpression(STAR_OP,NULL,NULL,NULL);
#else
		if (i>d) 
			node= new SgExpression(STAR_OP,NULL,NULL,NULL);
		else
			node= new SgExpression(BLOCK_OP,NULL,NULL,NULL);
#endif
			format_list = new SgExpression(EXPR_LIST,node,format_list,NULL);
		}
		if (dist_format)
		    dist_format->setRhs(format_list);
		else
		    dist_format = new SgExpression(DIST_FORMAT_OP,NULL,format_list,NULL);
		dist_op->setLhs(dist_format);
	    }
	}

	// or if dist_target is missed
	if (dist_target == NULL)
	{
	    SgStatement *hedr_stmt = s->getScopeForDeclare();
		// got the desire dimension, don't consider the * in dist_format
	    int dim=0;
	    for (SgExpression *expr_list=dist_format->rhs() ; expr_list ;
			expr_list=expr_list->rhs())
		if (expr_list->lhs()->variant() != STAR_OP) dim++;

		// get the processors if possible !
//	    if ( num_dims != -1) dim=num_dims;		// if argument define dimension
	    SgSymbol *procs=get_dim_processor(hedr_stmt,dim);
	    if (procs == NULL)
	    {
		printf("Can not handled to devide the %d procesors in %d dims ,aborted !\n",num_procs,dim);
                exit(1);
	    }

	    SgExpression *ref=new SgVarRefExp(procs);
	    dist_target = new SgExpression(DIST_TARGET_OP , NULL, ref,NULL);
	    dist_op->setRhs(dist_target);
	}
}


/****************************************************************************
*****************************************************************************/

void realign_son(SgStatement *s)
{
   SgSymbol *symb = s->expr(0)->lhs()->symbol();
   if (symb->numberOfAttributes(Align_Son) == 0) return;

   SgSymbol *sym_son=(SgSymbol *)symb->attributeValue(0,Align_Son);
   while (sym_son) {
      SgSymbol *next_son = (SgSymbol *)sym_son->attributeValue(0,Align_Son);
      if (next_son == NULL) break;
      sym_son = next_son;
   }
   while (sym_son != symb) {
printf("sym_son:%s\n", sym_son->identifier());
      SgStatement *align = (SgStatement *)sym_son->attributeValue(0,Align_Stmt);
      SgStatement *realign = align->copyPtr();
      realign->thebif->variant = REALIGN_STMT;
      s->insertStmtAfter(*realign, *s->controlParent());
      sym_son = (SgSymbol *)sym_son->attributeValue(0,Align_Parent);
   }
}




